#!/usr/bin/env python3
# coding=utf-8

from . import keys
from . import render
from . import search
from . import thumbnails
from .consts import STDSCR  # noqa: F401
from .keys import keys as k
from time import sleep
import curses


def set_curses_start_defaults():
    """Set curses start defaults."""
    curses.use_default_colors()
    curses.curs_set(0)  # Turn off cursor


def run(stdscr):
    global STDSCR
    STDSCR = stdscr  # override global STDSCR by the stdscr from wrapper
    page_dict = render.Tabs().fpagedict()  # last used tab/page
    page = render.Page(page_dict)

    set_curses_start_defaults()

    def redraw():
        """Reinitialize variables & redraw everything."""
        thumbnails.Draw().finish()
        STDSCR.clear()
        h, w = STDSCR.getmaxyx()
        if h < 3 or w < 3:
            return
        page.draw()
        thumbnails.Draw().start()

    redraw()  # draw once just before the loop start

    # Infinite loop to read every key press.
    while True:
        c = STDSCR.get_wch()
        if isinstance(c, int) and c == curses.KEY_RESIZE:  # terminal resize event
            # fix: handle crazy multiple repeated window resizing initiated by the user
            # NOTE: this introduces slight redraw delay after resize but fixes crashes
            while True:
                sleep(.25)
                nlines, ncols = STDSCR.getmaxyx()
                if curses.is_term_resized(nlines, ncols):
                    sleep(.75)
                    continue
                else:
                    break
            # redraw & start loop again without further more complex execution
            redraw()
            continue
        c = str(c)  # convert character to string
        h, w = STDSCR.getmaxyx()
        # Show last pressed key chars at the bottom-right corner.
        STDSCR.insstr(h - 1, w - 4, c)
        if c == k.get("quit"):
            break
        if c == k.get("redraw"):
            page = render.Page(page_dict)
            redraw()
            continue
        if c == k.get("full_title"):
            STDSCR.clear()
            fbox = render.Boxes.drawn_boxes[0]
            # toggle full title drawing
            if not fbox.fulltitle:
                page.draw(fulltitle=True)
            else:
                page.draw()
            continue
        if c == k.get("tab_find"):
            page_dict = render.Tabs().find_tab()
            page = render.Page(page_dict)
            redraw()
            continue
        if c == k.get("tab_add"):
            page_dict = search.select_page(page_dict)
            page = render.Page(page_dict)
            redraw()
            continue
        if c == k.get("tab_delete"):
            page_dict = render.Tabs().delete_tab()
            page = render.Page(page_dict)
            redraw()
            continue
        if c == k.get("tab_prev") or c == k.get("tab_next"):
            if c == k.get("tab_next"):
                page_dict = render.Tabs().next_tab()
            else:
                page_dict = render.Tabs().prev_tab()
            page = render.Page(page_dict)
            redraw()
            continue
        if keys.hints(c):
            # clear possible fulltitle str
            # hide previously shown hints etc.
            STDSCR.clear()
            page.draw()
            continue
        if keys.yank(c):
            continue
        if keys.scroll(c, page.draw):
            redraw()
            continue
    thumbnails.Draw().finish()
    sleep(0.3)


def main():
    curses.wrapper(run)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# coding=utf-8

from . import data
from . import render
from .clip import clip
from .conf import key as ck
from .notify import notify

keys = {
    "quit": ck("quit"),
    "redraw": ck("redraw"),
    "full_title": ck("full_title"),
    "tab_find": ck("tab_find"),
    "tab_add": ck("tab_add"),
    "tab_delete": ck("tab_delete"),
    "tab_next": ck("tab_next"),
    "tab_prev": ck("tab_prev"),
    "yank_urls": ck("yank_urls"),
    "yank_urls_page": ck("yank_urls_page"),
}

hint_keys = {
    "hint_clip_url": ck("hint_clip_url"),
    "hint_open_stream": ck("hint_open_stream"),
    "hint_open_extra": ck("hint_open_extra"),
    "hint_open_video": ck("hint_open_video"),
}

scroll_keys = {
    "scroll_top": ck("scroll_top"),
    "scroll_bot": ck("scroll_bot"),
    "scroll_up": ck("scroll_up"),
    "scroll_down": ck("scroll_down"),
    "scroll_up_page": ck("scroll_up_page"),
    "scroll_down_page": ck("scroll_down_page")
}


def scroll(c, page_draw):
    """Scroll page."""
    if c in scroll_keys.values():
        grid = page_draw()
        if c == scroll_keys.get("scroll_down"):
            grid.shift_index("down")
        elif c == scroll_keys.get("scroll_up"):
            grid.shift_index("up")
        elif c == scroll_keys.get("scroll_down_page"):
            grid.shift_index("down", page=True)
        elif c == scroll_keys.get("scroll_up_page"):
            grid.shift_index("up", page=True)
        elif c == scroll_keys.get("scroll_top"):
            grid.shift_index("top")
        elif c == scroll_keys.get("scroll_bot"):
            grid.shift_index("bot")
        return True
    return False


def hints(c):
    """Show hints, and make some action based on key and hint."""
    if c in hint_keys.values():
        hints = render.Hints()
        hint = hints.show_hints_boxes()
        if c == hint_keys.get("hint_clip_url"):
            hints.copy_url(hint)
        else:
            if c == hint_keys.get("hint_open_stream"):
                type = "stream"
            elif c == hint_keys.get("hint_open_video"):
                type = "video"
            elif c == hint_keys.get("hint_open_extra"):
                type = "extra"
            else:
                type = "stream"
            hints.open_url(hint, type)
        return True
    else:
        return False


def yank_urls(full_page=False):
    """Yank urls of visible boxes or all urls of the page."""
    urls = ""
    if full_page:
        page_dict = render.Tabs().fpagedict()  # current tab/page
        json_data = data.page_data(page_dict)
        if "url" in json_data["data"][0]:
            page_urls = data.get_entries(json_data, "url")
            for url in page_urls:
                urls += f"{url}\n"
    else:
        for box in render.Boxes.drawn_boxes:
            urls += f"{box.url}\n"
    if urls:
        clip(urls)
    else:
        notify("This page does not have 'url' entries in json data.")


def yank(c):
    if c == keys.get("yank_urls") or c == keys.get("yank_urls_page"):
        if c == keys.get("yank_urls"):
            yank_urls()
        elif c == keys.get("yank_urls_page"):
            yank_urls(full_page=True)
        return True
    else:
        return False



#!/usr/bin/env python3
# coding=utf-8

from . import conf
from . import open_cmd
from . import pages
from . import search
from . import utils
from .clip import clip
from .consts import STDSCR
from .iselect import iselect
from .thumbnails import container_size
from ast import literal_eval
from itertools import islice
from time import sleep
import curses
import re


class Hints:
    hint_chars = str(conf.setting("hint_chars"))
    active_hints = []

    def total(self, items) -> tuple[int, int]:
        """Return (total_seq, hint_length) based on hint_chars,
        individual hint_length formula and len of items.
        total_seq = number of possible sequences,
        hint_length = [1-3] length of each hint.
        """
        hcl = len(self.hint_chars)
        sqr = hcl ** 2
        if hcl >= len(items):
            hint_length = 1
            total_seq = hcl
        elif sqr >= len(items):
            hint_length = 2
            total_seq = sqr
        else:
            hint_length = 3
            total_seq = sqr * 2 - hcl
        return total_seq, hint_length

    def shorten_uniq_seq(self, out_seq):
        """Shorten all unique hint sequences and insert at the beginning."""
        tmp_seq = out_seq.copy()
        # shorten all sequences by one character
        for i, seq in enumerate(tmp_seq):
            wlc = seq[:-1]  # seq without last char
            tmp_seq.pop(i)
            tmp_seq.insert(i, wlc)
        # remove all elements that occurs more than once
        seq_set = set(tmp_seq)  # set() for less loop iterations
        for seq in seq_set:
            occurs = tmp_seq.count(seq)  # the number of times an element occurs
            if occurs > 1:
                while seq in tmp_seq:
                    tmp_seq.remove(seq)
        if not tmp_seq:  # short unique sequences not found
            return out_seq
        # each letter associated with its position index in self.hint_chars
        order = {}
        for i, c in enumerate(self.hint_chars):
            order[c] = i + 3
        # compute the seq score by the order in which the letters appear in the sequence
        seq_score = {}
        for seq in tmp_seq:
            s1 = order[seq[0]]
            s2 = 0
            if len(seq) > 1:
                s2 = order[seq[1]]
            score = s1 * 3 + s2
            seq_score[seq] = score
        # dict of sequences sorted by the sequence score
        sorted_by_score = dict(sorted(seq_score.items(), key=lambda x: x[1]))
        sorted_short_seq = list(sorted_by_score.keys())
        sorted_short_seq.reverse()  # reverse() => we insert at the beginning
        # replace original seq by the shorter sequence as it occurs only once
        for sseq in sorted_short_seq:
            # original long seq found by the short seq
            llseq = [s for s in out_seq if re.search(f"^{sseq}.", s)]
            lseq = str(llseq[0])
            if lseq and lseq in out_seq:
                out_seq.remove(lseq)
            # insert all short sequences at the beginning
            out_seq.insert(0, sseq)
        return out_seq

    def gen_hint_seq(self, items) -> list:
        """Generate from hint_chars list of unique sequences."""
        _, hint_length = self.total(items)

        # one letter length hints
        if hint_length == 1:
            return list(self.hint_chars)[:len(items)]

        # simple repeated values of hint_length
        repeated = []
        for c in self.hint_chars:
            # nn ee oo ... (if length_chars=2)
            repeated.append(c * hint_length)

        # make unique combinations of letters in strict order
        # generates sequence of 2 or 3 letter length hints
        combinations = []
        for r in repeated:
            new_seq = ""
            for ci in range(hint_length, 1, -1):
                pi = ci - 1
                for c in self.hint_chars:
                    new_seq = r[:pi] + c + r[ci:]
                    if new_seq in repeated:
                        continue  # skip
                    if new_seq in combinations:
                        continue  # skip
                    combinations.append(new_seq)

        hint_sequences = []
        hint_sequences.extend(repeated)
        hint_sequences.extend(combinations)
        # limit by the number of sequences that is enough for all items
        out_seq = hint_sequences[:len(items)]
        # NOTE: short seq are more convenient to type
        out_seq = self.shorten_uniq_seq(out_seq)
        # limit the number of sequences, strictly after shortening! (just in case)
        if len(out_seq) > len(items):
            return out_seq[:len(items)]
        else:
            return out_seq

    def hint(self, items: list) -> list:
        """Return hint sequences for items."""
        hints = self.gen_hint_seq(items)
        self.active_hints = hints
        return hints

    def find_seq(self, hints) -> str:
        """Input characters until only one hint sequence is found."""
        cinput = ""
        select = hints
        while len(select) > 1:
            c = str(STDSCR.get_wch())
            cinput += c
            select = [s for s in select if re.search(f"^{cinput}", s)]
        if not select:
            return ""
        if len(select) != 1:
            raise ValueError(f"len:({len(select)}) Only one item should be in the list:\n{select}")
        return str(select[0])

    def show_hints_boxes(self):
        """Show hints for visible/drawn boxes."""
        boxes = Boxes.drawn_boxes
        hints = self.hint(boxes)
        for box, hint in zip(boxes, hints):
            box.hint = hint
            box.show_hint()
        return self.find_seq(hints)

    def get_box_attr_hint(self, hint, attr):
        """return attribute value of box object found by the hint."""
        boxes = Boxes.drawn_boxes
        if not hasattr(boxes[0], attr):
            raise AttributeError(f"'{attr}' -> {boxes[0]} does not have such attribute!")
        for box in boxes:
            if getattr(box, "hint") == hint:
                return getattr(box, attr)
        raise Exception(f"value of ATTR:'{attr}' by the HINT:'{hint}' not found!")

    def copy_url(self, hint):
        if not hint:
            return
        value = self.get_box_attr_hint(hint, "url")
        clip(value)

    def open_url(self, hint, type):
        if not hint:
            return
        value = self.get_box_attr_hint(hint, "url")
        open_cmd.open_url(value, type)


class Boxes:
    """Operate on list of Boxes"""
    boxlist = []
    thmblist = []
    drawn_boxes = []

    def add(self, obj):
        """Add box object to list."""
        self.boxlist.append(obj)

    def add_thmb(self, obj):
        """Add thumbnail ue params to list."""
        self.thmblist.append(obj)

    def draw(self, parent, grid, fulltitle=False):
        """Draw boxes."""
        self.drawn_boxes.clear()
        stop = len(grid.coordinates())
        for box in islice(self.boxlist, stop):
            if fulltitle:
                box.fulltitle = True
            box.draw(parent)
            self.drawn_boxes.append(box)
        parent.refresh()
        self.boxlist.clear()


class Box:
    """Box with info about the stream/video inside the Grid."""
    w, h = container_size()
    last = h - 2  # last line of the box

    def __init__(self, user_login, user_name, title, category, x, y):
        self.user_login = user_login  # for composing url
        self.user_name = user_name
        self.title = utils.strclean(title)
        self.category = category
        self.x = x
        self.y = y
        self.url = f"https://www.twitch.tv/{self.user_login}"
        self.hint = ""
        self.img_path = ""
        self.viewers = ""
        self.duration = ""
        self.fulltitle = False

    def draw(self, parent):
        """Draw Box."""
        win = parent.derwin(self.h, self.w, self.y, self.x)
        win.addnstr(self.last, 0, f"{self.category}", self.w)
        if self.duration:
            duration = f"[{self.duration}]"
            rside = self.w - len(duration)
            win.addnstr(self.last - 1, 0, f"{self.user_name}", rside, curses.A_BOLD)
            win.addstr(self.last - 1, rside, duration)
        else:
            win.addnstr(self.last - 1, 0, f"{self.user_name}", self.w, curses.A_BOLD)
        if self.viewers:
            viewers = f" {self.viewers}"
            rside = self.w - len(viewers)
            win.addstr(self.last, rside, viewers, curses.A_BOLD)
        if self.fulltitle:
            max_len = int(self.w * 3)  # 3 box widths (lines)
            title = utils.word_wrap_title(self.title, self.w, max_len)
            try:
                win.addnstr(self.last - 2, 0, title, max_len)
            except Exception:
                win.box()
        else:
            title = utils.strtoolong(self.title, self.w)
            win.addnstr(self.last - 2, 0, title, self.w)

    def show_hint(self):
        """Create window with hint character."""
        if self.hint:  # if hint not empty -> show hint
            curses.init_pair(1, curses.COLOR_WHITE, curses.COLOR_RED)
            if len(self.hint) == 1:
                hint = f" {self.hint} "
            elif len(self.hint) == 2:
                hint = f" {self.hint}"
            else:
                hint = f"{self.hint}"
            lh = len(hint)
            win = curses.newwin(1, lh + 1, self.y + self.h - 1, self.x + self.w - lh)
            win.addstr(hint, curses.color_pair(1))
            win.refresh()


class Grid:
    """Grid of boxes inside the Window."""
    w, h = container_size()

    def __init__(self, key_list: list, page_name: str):
        self.key_list = key_list
        self.page_name = page_name
        self.__ba = BodyArea()
        self.area_cols = self.__ba.cols
        self.area_rows = self.__ba.rows
        self.key_start_index = self.index()
        self.coords = self.coordinates()

    def capacity(self) -> tuple[int, int, int]:
        """Return - how many boxes can fit in: (cols, rows, total)."""
        cols = self.area_cols // self.w
        rows = self.area_rows // self.h
        total = cols * rows
        return cols, rows, total

    def spacing(self, cols, rows) -> tuple[int, int]:
        """Calculate even spacing between grid elements.
        returns spacing: cols, rows.
        """
        if cols < 1:
            c = 0
        else:
            c = int(self.area_cols - self.w * cols) // cols
        if rows < 1:
            r = 0
        else:
            r = int(self.area_rows - self.h * rows) // rows
        return c, r

    def index(self, start_index="") -> int:
        """Set/Get initial grid index."""
        if str(start_index):  # str -> to check if not empty (even 0 value)
            index = int(start_index)
            conf.tmp_set("grid_index", index, self.page_name)
        else:
            index = int(conf.tmp_get("grid_index", 0, self.page_name))
            if not index or index < 0:
                index = 0
                conf.tmp_set("grid_index", index, self.page_name)
        return index

    def shift_index(self, dir="down", page=False) -> int:
        """Shift value of the key start index."""
        cols, _, total = self.capacity()
        elems_total = len(self.key_list)
        remainder = elems_total % cols
        if remainder == 0:
            end_of_page = elems_total - total
        else:
            end_of_page = elems_total - total - remainder + cols
        if elems_total <= total:
            start_index = 0
        else:
            grid_index = self.index()
            if dir == "top":
                start_index = 0
            elif dir == "bot":
                start_index = end_of_page
            elif page:
                if dir == "down":
                    start_index = grid_index + total
                else:
                    start_index = grid_index - total
            else:
                if dir == "down":
                    start_index = grid_index + cols
                else:
                    start_index = grid_index - cols
            if start_index < 0:
                start_index = 0
            elif start_index > end_of_page:
                start_index = end_of_page
        self.index(start_index)
        return start_index

    def coordinates(self) -> dict:
        """Return dict with: tuple(X, Y) values where each key_list element is the key."""
        initial_x, initial_y = 0, 0
        cols, rows, total = self.capacity()
        total += self.key_start_index  # for scrolling
        scols, srows = self.spacing(cols, rows)
        # for more even spacing from both sides
        sc = scols // 2
        sr = srows // 2
        x = initial_x + sc
        y = initial_y + sr
        current_col = 1
        coordinates = {}
        for key in islice(self.key_list, self.key_start_index, total):
            if cols > 2:
                x += sc
            coordinates[key] = (x, y)
            if current_col < cols:
                current_col += 1
                x += sc + self.w
            else:
                current_col = 1
                x = initial_x + sc
                y += sr + self.h
        return coordinates


class Tabs:
    """Tabs."""
    try:  # list of tab names
        tabs = literal_eval(conf.tmp_get("tabs", [], "TABS"))
    except ValueError:  # handle literal_eval() error with empty list
        tabs = []

    def curtab(self):
        """Get current page name, or 'Following Live' as fallback."""
        return conf.tmp_get("current_page_name", "Following Live", "TABS")

    def fpagedict(self, tab_name="") -> dict:
        """Find and return page dict by the tab name or for current tab."""
        if not tab_name:  # return page_dict of current tab/page
            page_dict_str = conf.tmp_get("page_dict", "", self.curtab())
        else:
            page_dict_str = conf.tmp_get("page_dict", self.curtab(), tab_name)
        if not page_dict_str or page_dict_str == "Following Live":
            return search.following_live()
        try:
            page_dict = literal_eval(page_dict_str)
        except Exception as e:
            raise ValueError(f"page_dict_str: '{page_dict_str}'\n{e}")
        return page_dict

    def add_tab(self, page_name):
        current_page_name = self.curtab()
        # don't add the same tab twice
        if page_name not in self.tabs:
            if not self.tabs or current_page_name not in self.tabs:
                self.tabs.append(current_page_name)
                if page_name not in self.tabs:
                    self.tabs.append(page_name)
            else:
                # find index of current page name and insert new tab after that index
                cindex = self.tabs.index(current_page_name)
                nindex = cindex + 1
                self.tabs.insert(nindex, page_name)
            conf.tmp_set("tabs", self.tabs, "TABS")

    def delete_tab(self):
        """Delete current tab and return page_dict of the previous tab."""
        ptabname = self.prev_tab(tab_name=True)
        self.tabs.remove(self.curtab())
        conf.tmp_set("tabs", self.tabs, "TABS")
        return self.fpagedict(ptabname)

    def find_tab(self) -> dict:
        """Find and return page dict of selected tab."""
        mulstr = "\n".join(self.tabs)  # each list element on it's own line
        tabname = iselect(mulstr, 130)
        # handle cancel of the command
        if tabname == 130:
            # fallback to current tab
            return self.fpagedict(self.curtab())
        return self.fpagedict(tabname)

    def next_tab(self, tab_name=False):
        """Return page_dict for the next tab name (carousel) or simply tab_name."""
        cindex = self.tabs.index(self.curtab())
        nindex = cindex + 1
        if nindex > len(self.tabs) - 1:
            ntabname = self.tabs[0]
        else:
            ntabname = self.tabs[nindex]
        if tab_name:
            return ntabname
        else:
            return self.fpagedict(ntabname)

    def prev_tab(self, tab_name=False):
        """Return page_dict for the prev tab name (carousel) or simply tab_name."""
        cindex = self.tabs.index(self.curtab())
        pindex = cindex - 1
        ptabname = self.tabs[pindex]
        if tab_name:
            return ptabname
        else:
            return self.fpagedict(ptabname)


class Page:
    """Page which renders everything."""
    HEADER_H = pages.Pages.HEADER_H

    def __init__(self, page_dict):
        self.pages_class = pages.Pages(page_dict)
        self.page_name = self.pages_class.page_name
        self.grid_func = self.pages_class.grid_func
        self.loaded = False

    @utils.background
    def loading(self):
        """Simple animation to show that currently something is being done."""
        chars = "-\\|/"  # animation chars
        win = curses.newwin(1, 1, 0, 0)
        for _ in range(25):
            for c in chars:
                win.insstr(c)
                win.refresh()
                sleep(.1)
                if self.loaded:
                    break
            if self.loaded:
                break
        win.erase()
        win.refresh()

    def draw_header(self):
        """Draw page header."""
        indent = 2  # indent from side
        indent_between = " "
        separator = "|"  # separator between tabs
        between_tabs = indent_between + separator + indent_between
        logo = "[twitchez]"
        c_page = self.page_name  # current page name
        _, w = STDSCR.getmaxyx()
        head = STDSCR.derwin(self.HEADER_H - 1, w, 0, 0)
        other_tabs = ""
        # tab order where current page is always first in list (to look as carousel)
        taborder = []
        tabs = Tabs.tabs
        cpni = tabs.index(c_page)
        taborder.extend(tabs[cpni:])
        taborder.extend(tabs[:cpni])
        for tab in taborder:
            if tab == c_page:
                continue  # skip current tab
            # indent_between with separator for each additional tab page
            other_tabs += between_tabs + tab
        icp = indent + len(c_page)  # width of current page with indent
        all_tabs = indent + len(c_page + other_tabs)
        if w > all_tabs:  # if we can fit all tabs
            head.addnstr(0, indent, c_page, len(c_page), curses.A_REVERSE)  # current Tab page
            head.addnstr(0, icp, other_tabs, len(other_tabs))
            if w > all_tabs + len(logo) + 1:  # if window have enough width for logo
                wllimit = w - len(logo) - indent
                head.addnstr(0, wllimit, logo, len(logo), curses.A_BOLD)
        else:
            # crop tabs visually that do not fit into the window
            # > character signalize about cropping (existence of other tabs)
            if w > indent:
                wclimit = w - indent - 1
                c_page = c_page[:wclimit - 1] + ">"
                head.addnstr(0, indent, c_page, wclimit, curses.A_REVERSE)
            if w > icp:
                wolimit = w - icp - 1
                other_tabs = other_tabs[:wolimit - 1] + ">"
                head.addnstr(0, icp, other_tabs, wolimit)
        head.refresh()
        return head

    def draw_body(self, grid, fulltitle=False):
        """Draw page body."""
        body = BodyArea().window()
        if fulltitle:
            Boxes().draw(body, grid, fulltitle)
        else:
            Boxes().draw(body, grid)
        return body

    def draw(self, fulltitle=False):
        """return grid and draw full page."""
        self.loading()
        grid = self.grid_func()
        self.draw_body(grid, fulltitle)
        self.draw_header()
        self.loaded = True  # finish loading animation
        return grid


class BodyArea:
    """Body area size of the window in the terminal cells."""
    HEADER_H = pages.Pages.HEADER_H

    def __init__(self):
        self.rows, self.cols = STDSCR.getmaxyx()
        self.rows = self.rows - self.HEADER_H

    def window(self):
        """Create & return body window."""
        return STDSCR.derwin(self.rows, self.cols, self.HEADER_H, 0)



#!/usr/bin/env python3
# coding=utf-8

from os import environ
from pathlib import Path
from tempfile import gettempdir


def project_root(*args) -> Path:
    """Simply return project_root or compose path from args."""
    if not args:
        return Path(__file__).parent.parent
    else:
        return Path(Path(__file__).parent.parent, *args)


def get_cache_dir() -> Path:
    """Check ENV variables, create cache dir and return it's path."""
    dirname = "twitchez"
    if "TWITCHEZ_CACHE_DIR" in environ:
        cache_home = environ["TWITCHEZ_CACHE_DIR"]
    elif "XDG_CACHE_HOME" in environ:
        cache_home = environ["XDG_CACHE_HOME"]
    else:
        cache_home = Path(Path.home(), ".cache")
    cache_dir = Path(cache_home, dirname)
    # create cache_dir if not exist
    Path(cache_dir).mkdir(parents=True, exist_ok=True)
    return cache_dir


def get_tmp_dir(*subdirs) -> Path:
    """Return path to tmp dir and create optional subdirs if they doesn't already exist."""
    dirname = "twitchez"
    if not subdirs:
        tmp_dir_path = Path(gettempdir(), dirname)
    else:
        tmp_dir_path = Path(gettempdir(), dirname, *subdirs)
    Path(tmp_dir_path).mkdir(parents=True, exist_ok=True)
    return tmp_dir_path


def get_user_conf_dir() -> Path:
    """Check ENV variables, get user config dir and return it's path."""
    dirname = "twitchez"
    if "XDG_CONFIG_HOME" in environ:
        config_home = environ["XDG_CONFIG_HOME"]
    else:
        config_home = Path(Path.home(), ".config")
    config_dir = Path(config_home, dirname)
    return config_dir



#!/usr/bin/env python3
# coding=utf-8

import curses

# NOTE: should be always overridden by the stdscr from wrapper
# defined here just to be able to see actual curses funcs of stdscr
# and clarify variable type for: interpreter, diagnostic messages, developer etc.
STDSCR = curses.initscr()



#!/usr/bin/env python3
# coding=utf-8

__version__ = "0.0.1a"
__license__ = "GPLv3"
__description__ = "TUI client for twitch with thumbnails"
__url_repository__ = "https://github.com/WANDEX/twitchez"
__url_bug_reports__ = "https://github.com/WANDEX/twitchez/issues"
__url_project__ = __url_repository__
__author__ = "WANDEX"



#!/usr/bin/env python3
# coding=utf-8

from . import fs
from configparser import ConfigParser
from pathlib import Path

glob_conf = fs.project_root("config", "default.conf")
user_conf = fs.project_root(fs.get_user_conf_dir(), "config.conf")

glob_keys = fs.project_root("config", "keys.conf")
user_keys = fs.project_root(fs.get_user_conf_dir(), "keys.conf")

config = ConfigParser()
config.read(glob_conf)
config.read(user_conf)  # user config takes precedence over global default config

keymap = ConfigParser()
keymap.read(glob_keys)
keymap.read(user_keys)  # user config takes precedence over global default config

temp_vars = Path(fs.get_tmp_dir(), "vars")
temp = ConfigParser()


def setting(keyname):
    found = config.get("GENERAL", keyname)
    return found


def key(keyname):
    found = keymap.get("GENERAL", keyname, fallback="")
    return found


def tmp_set(option, value, section="GENERAL"):
    """Set tmp variable value."""
    temp.read(temp_vars)
    if not temp.has_section(section):
        temp.add_section(section)
    temp.set(str(section), str(option), str(value))
    with open(temp_vars, "w") as file:
        temp.write(file, space_around_delimiters=False)


def tmp_get(keyname, fallback, section="GENERAL"):
    """Get tmp variable value."""
    temp.read(temp_vars)
    if not temp.has_section(section):
        temp.add_section(section)
    if fallback or not temp.has_option(section, keyname):
        found = temp.get(section, keyname, fallback=fallback)
    else:
        found = temp.get(section, keyname)
    return found



#!/usr/bin/env python3
# coding=utf-8

from setuptools import find_packages, setup
import twitchez

with open("README.md", "r") as f:
    long_description = f.read()
with open("requirements.txt", "r") as f:
    requirements = [line.strip() for line in f]

setup(
    name="twitchez",
    version=twitchez.__version__,
    license=twitchez.__license__,
    url=twitchez.__url_project__,
    author=twitchez.__author__,
    description=twitchez.__description__,
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    python_requires=">=3.6",
    install_requires=requirements,
    extras_require={
        "thumbnails": ["ueberzug"],  # (X11 only)
    },
    package_data={
        "": ["config/*.conf", "config/blank.jpg"],
    },
    entry_points={
        "console_scripts": ["twitchez=twitchez.init:main"]
    },
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3",
        "Environment :: Console :: Curses",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Development Status :: 2 - Pre-Alpha",
    ],
    keywords="twitch TUI terminal ui curses client thumbnail image twitch.tv",
    project_urls={
        "Bug Reports": twitchez.__url_bug_reports__,
        "Source": twitchez.__url_repository__,
    },
)



#!/usr/bin/env python3
# coding=utf-8

from . import conf
from datetime import datetime
from difflib import SequenceMatcher
from os.path import getmtime
from re import compile
from threading import Thread
import textwrap
import time


# visible length of one emoji in terminal cells
EMOJI_CELLS = int(conf.setting("emoji_cells"))

EMOJI_PATTERN = compile(
    "["
    "\U0001F1E0-\U0001F1FF"  # flags (iOS)
    "\U0001F300-\U0001F5FF"  # symbols & pictographs
    "\U0001F600-\U0001F64F"  # emoticons
    "\U0001F680-\U0001F6FF"  # transport & map symbols
    "\U0001F700-\U0001F77F"  # alchemical symbols
    "\U0001F780-\U0001F7FF"  # Geometric Shapes Extended
    "\U0001F800-\U0001F8FF"  # Supplemental Arrows-C
    "\U0001F900-\U0001F9FF"  # Supplemental Symbols and Pictographs
    "\U0001FA00-\U0001FA6F"  # Chess Symbols
    "\U0001FA70-\U0001FAFF"  # Symbols and Pictographs Extended-A
    "\U00002702-\U000027B0"  # Dingbats
    "\U000024C2-\U0001F251"
    "]+"
)


def tryencoding(string: str) -> str:
    """Return string in default encoding or
    if not printable -> try to re-encode into utf-16."""
    if not string.isprintable():
        try:
            string = string.encode('utf-16', 'surrogatepass').decode("utf-16", "ignore")
        except (UnicodeEncodeError, UnicodeDecodeError) as e:
            string = str(e)
    return string


def demojize(str: str) -> str:
    """Return string without emojis."""
    return EMOJI_PATTERN.sub(r'', str)


def emoji_count(str: str) -> int:
    """Returns the count of emojis in a string."""
    return len(str) - len(demojize(str))


def tlen(str: str) -> int:
    """Return len of str respecting emoji visible length in terminal cells.
    EMOJI_CELLS: visible length of one emoji in terminal cells.
    """
    if EMOJI_CELLS < 2:
        return len(str)
    else:
        return EMOJI_CELLS * emoji_count(str) + len(demojize(str))


def secs_since_mtime(path):
    """time_now - target_mtime = int(secs)."""
    return int(time.time() - getmtime(path))


def replace_pattern_in_all(inputlist, oldstr, newstr) -> list:
    """Replace oldstr with newstr in all items from a list."""
    outputlist = []
    for e in inputlist:
        outputlist.append(str(e).replace(oldstr, newstr))
    return outputlist


def add_str_to_list(input_list, string) -> list:
    """Add string to the end of all elements in a list."""
    outputlist = [e + str(string) for e in input_list]
    return outputlist


def insert_to_all(list, string, opt_sep="") -> list:
    """ Insert the string at the beginning of all items in a list. """
    string = str(string)
    if opt_sep:
        string = f"{string}{opt_sep}"
    string += '% s'
    list = [string % i for i in list]
    return list


def strws(str: str) -> str:
    """Return a str with whitespace characters replaced by '_'."""
    return str.strip().replace(" ", "_")


def strclean(str: str) -> str:
    """return slightly cleaner string."""
    # remove unneeded characters from string
    s = str.replace("\n", " ").replace("\t", " ")
    # replace repeating whitespaces by single whitespace
    s = ' '.join(s.split())
    s = s.strip()
    return s


def strtoolong(str: str, width: int, indicator="..") -> str:
    """Return str slice of width with indicator at the end.
    (to show that the string cannot fit completely in width)
    """
    if tlen(str) > width:
        str_fit_in_width = str[:width]
        # visible width in terminal cells that str occupies
        terminal_cells = tlen(str_fit_in_width)
        if terminal_cells > width:
            ec = emoji_count(str_fit_in_width)
            cut = ec + len(indicator)
            out_str = str_fit_in_width[:-cut] + indicator
        else:
            out_str = str_fit_in_width[:-len(indicator)] + indicator
        return out_str
    else:
        return str


def word_wrap_title(string: str, width: int, max_len: int, max_lines=3) -> str:
    """Word wrap title string."""
    string = strclean(string)
    if tlen(string) <= width:
        return string
    title_lines = textwrap.wrap(
        string, width, max_lines=max_lines,
        expand_tabs=False, replace_whitespace=True,
        break_long_words=True, break_on_hyphens=True, drop_whitespace=True
    )
    out_str = ""
    cline = 0
    for line in title_lines:
        cline += 1
        if len(line) == width:
            out_str += line
        else:
            out_str += f"{line}\n"
    # limit string len
    if len(out_str) > max_len:
        out_str = out_str[:max_len]
    # add mask only if length of last line met condition
    if len(title_lines[-1]) < width // 2:
        mask = "  "  # mask to differentiate from underlying text
        out_str = out_str[:-len(mask) + 1] + mask
    return out_str


def sdate(isodate: str) -> str:
    """Take iso date str and return shorten date str."""
    # remove Z character from default twitch date (2021-12-08T11:43:43Z)
    idate = isodate.replace("Z", "")
    vdate = datetime.fromisoformat(idate).isoformat(' ', 'minutes')
    today = datetime.today().isoformat(' ', 'minutes')
    current_year = today[:4]
    if current_year not in vdate:
        pattern = vdate[-6:]  # cut off only time
    else:
        sm = SequenceMatcher(None, vdate, today)
        match = sm.find_longest_match(0, len(vdate), 0, len(today))
        # longest common string between two
        pattern = vdate[match.a: match.a + match.size - 1]
    # remove pattern, cut leading '-' and strip whitespaces
    sdate = str(vdate).replace(pattern, "").strip("-").strip()
    return sdate


def duration(duration: str, simple=False, noprocessing=False) -> str:
    """Take twitch duration str and return duration with : as separators.
    Can optionally return a str without processing or with simple str processing.
    """
    if noprocessing:
        return duration
    if simple:
        # downside is very variable length of str and subjective ugliness of result.
        return duration.replace("h", ":").replace("m", ":").replace("s", ":").strip(":")
    # Don't see any real benefit of the following code over a silly simple one-liner :)
    # Result of the following algorithm are prettier, but also produces longer str.
    if "h" in duration:
        # extract hours from string
        H, _, _ = duration.partition("h")
        H = int(H.strip())
        # fix: if hours > 23 => put hours as simple str into format
        if H > 23:
            ifmt = f"{H}h%Mm%Ss"
            ofmt = f"{H}:%M:%S"
        else:
            ifmt = "%Hh%Mm%Ss"
            ofmt = "%H:%M:%S"
    elif "m" in duration:
        ifmt = "%Mm%Ss"
        ofmt = "%M:%S"
    else:
        return duration
    idur = datetime.strptime(duration, ifmt)
    odur = str(idur.strftime(ofmt))
    return odur


def background(func):
    """use @background decorator above the function to run in the background."""
    def background_func(*args, **kwargs):
        Thread(target=func, args=args, kwargs=kwargs).start()
    return background_func
